/* ifi.h
 * definition of the in-flight instruction class
 *
 * Each in-flight inst points to the icode of its source instruction.
 * The ifi accesses the physical registers in its containing thread
 */

#ifndef IFI_H
#define IFI_H


class inflight_inst {
public:

  inflight_inst(path_ptr mypath, icode_ptr mypicode);

  long &iREG(Arg _arg)
    { return pthread->Reg[args[_arg]].Int64; }

  double &iFP(Arg _arg)
    { return pthread->Reg[args[_arg]].Double; }

  long &iMapReg(Logical_RegNum lreg)
    { return ppath->pMapReg(lreg); }

  double &iMapFP(Logical_RegNum lfp)
    { return ppath->pMapFP(lfp); }

  void * addr2phys(aint_addr_t _vaddr, long *paccess)
    { return pthread->pprocess->addr2phys(_vaddr, paccess); }

  icode_ptr picode;
  thread_ptr pthread;
  path_ptr ppath;

  int iflags;

  Physical_RegNum args[MaxArgs];
  Physical_RegNum OldDest_Phys;
  Physical_RegNum NewDest_Phys;

  ulong fpcr;

  ulong fetchnum;    /* assigned in increasing order */

  state_t runstate;

  int signal;

  /* address computed by branch (cond, uncond, indirect) */
  aint_addr_t nextPC;  

  /* Get the current pc from picode */
  aint_addr_t get_pc() { return picode->addr; }

  /* for memops: */ 
  tb_entry *pte;
  aint_addr_t vaddr;
  aint_addr_t paddr;
  ulong data;

  int tid() { return pthread->tid; }

  /* data and methods for accessing full-empty bit vector */
  ulong *get_fe_vector() { return fe_vector; }
  ulong get_fe_mask() { return fe_mask; }

  /* The backend should not access these directly */
  ulong *fe_vector;
  ulong fe_mask;

  ulong bitmask;
  class stb_entry *stb_update;

  aint_logical committed;
  aint_logical done_memop;

  /* FIXME: is this being updated? Vega 05/29/97 */
  aint_logical BranchTaken;


  /* inflight_inst_ptr next;  // In fetch order. Put this in if needed */
  int iwpos;    // Position in the instwin circular buffer */

  int pos() { return iwpos; }
  void pos(int _pos) { iwpos = _pos; }

};

// These #defines are used by functions in op_asm.c and fpoper.c
#define REG(Rn) ifi->iREG(Rn)
#define FP(Rn) ifi->iFP(Rn)

#define MapReg(Rn) (ifi->iMapReg(Rn))
#define MapFP(Fn) (ifi->iMapFP(Fn))

#define IR2V(_picode) (_picode->addr)
#define IV2R(_pprocess, _aint_addr) (_pprocess->addr2iphys(_aint_addr, NULL))

#endif
