/* sim.h
 * prototypes of sim_ functions called from aint,
 * and AINT_ functions called from the simulator
 */

#ifndef AINT_SIM_H
#define AINT_SIM_H

// Must return correct value of extern optind.
extern int sim_init(int argc, char **argv, char **envp);

// Called when aint has problems parsing the command line
extern void sim_usage(int argc, char **argv);

// Called to transfer control to the backend
extern void sim_execute();

// Informs the backend when a new thread is created
extern void sim_thread_begin(int tid, aint_addr_t pc);

// Informs the backend when a thread dies
extern void sim_thread_end(int tid);

// Informs the backend when a thread is stalled (waiting)
extern void sim_thread_block(int tid);

// Informs the backend when an blocked thread can resume
extern void sim_thread_unblock(int tid);

// Called by aint before exit()
extern void sim_done(void);


// The AINT_* functions, called by the simulator

// Tells AINT to fetch from specified address into path
extern inflight_inst_ptr AINT_fetch(path_ptr ppath, aint_addr_t Addr);

// Issue the specified instruction. In case of memops, do address calculation
extern void AINT_issue(inflight_inst_ptr ifi);

// If (ifi->iflags & E_READ), load value from memory/store-buffer into reg
extern void AINT_do_read(inflight_inst_ptr ifi);

// Write value through to the store-buffer
extern void AINT_do_write(inflight_inst_ptr ifi);

// Retire instruction; free old mapping; flush store-buffer entry, if any
extern void AINT_commit(inflight_inst_ptr ifi);

// If this path has children, kill them. Then, roll-back from last inst
// in pipeline through ifi
extern void AINT_kill(inflight_inst_ptr ifi);

// Create a new root path in the specified thread
extern path_ptr AINT_create_path(int tid, aint_addr_t start_pc);

// Create a new path starting at start_pc, and make it a child of ifi->ppath.
extern path_ptr AINT_create_path(inflight_inst_ptr ifi, aint_addr_t start_pc);

// Create a path as the child of ppath
extern path_ptr AINT_create_path(path_ptr ppath, aint_addr_t start_pc);

// Recursively kill the path-tree rooted at ppath (depth-first search and kill)
extern void AINT_kill_path(path_ptr ppath);




#endif
