/* thread.h
 * definition of class thread
 */

#ifndef AINT_THREAD_H
#define AINT_THREAD_H

#include "aint_dllist.h"

class Store_Buffer;

class thread {
public:
  // For main thread in a leader (original) process
  thread(process_ptr);

  // For first thread in forked process
  thread(process_ptr, thread_ptr);

  // For a new thread in pthread_create
  thread(thread_ptr, aint_addr_t);

  int tid;
  process_ptr pprocess;

  RegType Reg[N_Physical_Regs];

  ulong InUse[(N_Physical_Regs+63)/64]; /* Need at least as many bits as registers */

  void Reset_InUse() { int i; for(i=0; i<(N_Physical_Regs+63)/64; i++) InUse[i] = 0ul; }

  int Is_InUse(Physical_RegNum n) { return (int) ((InUse[n>>6] >> (n&63)) & 1l); }

  void Set_InUse(Physical_RegNum n) { InUse[n>>6] |= (1l << (n&63)); }

  void Clr_InUse(Physical_RegNum n) {
    FreeRegStackTop++;
    FreeRegStack[FreeRegStackTop] = n;
    InUse[n>>6] &= ( ~(1l << (n&63)) );
  }

  Physical_RegNum ReMap(path_ptr ppath, Logical_RegNum reg);

  Physical_RegNum FreeRegStack[N_Physical_Regs]; // stack is simple and fast
  int FreeRegStackTop;

  Physical_RegNum RegMap[2*N_Logical_Regs];

  // Pointer to the main path
  path_ptr path_tree;

  icode_ptr picode;  /* used when thread state contains PC */
  long unique; /* Thread unique value used by rduniq and wruniq */

  aint_addr_t locked_addr;
  int lock_flag; /* Used by ldx_l and stx_c insts */

  Store_Buffer *StoreBuffer;

  long FetchCount;
  state_t runstate;
  int exitcode;

  // Thread-Specific Data items
  ulong TSD[AINT_pthread_keys_max];

  // Thread(s) waiting on us
  dllist Waiting_Threads;

  /* Points to the next thread created in this process - maybe should be a struct in process? */
  thread_ptr next_thread;
};


#endif
