/* pthread.c */

/* Implementation of AINT-pthread stubs that generate "fake" pthread syscalls
 * which are trapped by AINT.
 */


#include <stdio.h>

#include <aint_pthread.h>

/* Disable default definition of posix thread structures ?*/
/* #define POSIX_C_SOURCE 198001 */


#define CALLSYS 131
/* for call_pal */

#define DEBUG
#undef DEBUG

int __pthread_create(
		   pthread_t                *thread,
		   const pthread_attr_t     *attr,
		   void *                   (*start_routine)(void *),
		   void                     *arg)
{

    long callno = SYS_aint_pthread_create;
    long ret_val;
    pthread_t *newthread = thread;

    /* Copy the arguments into real registers */

#ifdef DEBUG
    printf("AINT pthread_create: making syscall %d\n", callno);
    fflush(stdout);
#endif

    asm(  "\n"
	  "bis %1, $31, $0   \n"
	  "bis %2, $31, $16  \n"
	  "bis %3, $31, $17  \n"
	  "bis %4, $31, $18  \n"
	  "bis %5, $31, $19  \n"
	  "call_pal      131 \n"
	  "bis $0, $31, %0   \n"
	  : "=r" (ret_val)
	  : "r" (callno), "r" (newthread), "r" (attr), "r" (start_routine), "r" (arg)
	  : "$16", "$17", "$18", "$19", "$0"
	  );

    return 0;

}



void __pthread_exit(void *value_ptr)
{

    long ret_val;
    long callno = SYS_aint_pthread_exit;

#ifdef DEBUG
  printf("AINT pthread_exit: making syscall %d\n", callno); fflush(stdout);
#endif

  asm( "bis %1, $31, $0 \n"
       "bis %2, $31, $16 \n"
       "call_pal     131 \n"
       "bis $0, $31, %0 \n"
       : "=r" (ret_val)
       : "r" (callno), "r" (value_ptr)
       : "$16", "$17", "$18", "$19", "$0"
       );
  /* return (ret_val); */
}

int __pthread_join (pthread_t thread, void **value_ptr)
{
    long ret_val;
    long callno = SYS_aint_pthread_join;

#ifdef DEBUG
  printf("AINT pthread_join: making syscall %d\n", callno); fflush(stdout);
#endif
  asm( "bis %1, $31, $0 \n"
       "bis %2, $31, $16 \n"
       "bis %3, $31, $17 \n"
       "call_pal     131 \n"
       "bis $0, $31, %0 \n"
       : "=r" (ret_val)
       : "r" (callno), "r" (thread), "r" (value_ptr)
       : "$16", "$17", "$18", "$19", "$0"
       );
  return (ret_val);

}

// Note: this is a new function name (from keycreate), and is not mangled
//  (by prefixing with __)

int pthread_key_create (pthread_key_t *key, void (*destructor)(void *))
{
    long ret_val;
    long callno = SYS_aint_pthread_key_create;

#ifdef DEBUG
  printf("AINT pthread_key_create: making syscall %d\n", callno); fflush(stdout);
#endif
  asm( "bis %1, $31, $0 \n"
       "bis %2, $31, $16 \n"
       "bis %3, $31, $17 \n"
       "call_pal     131 \n"
       "bis $0, $31, %0 \n"
       : "=r" (ret_val)
       : "r" (callno), "r" (key), "r" (destructor)
       : "$16", "$17", "$18", "$19", "$0"
       );
  return (ret_val);

}

int __pthread_setspecific (pthread_key_t key, void *value)
{
    long ret_val;
    long callno = SYS_aint_pthread_setspecific;

#ifdef DEBUG
  printf("AINT pthread_setspecific: making syscall %d\n", callno); fflush(stdout);
#endif
  asm( "bis %1, $31, $0 \n"
       "bis %2, $31, $16 \n"
       "bis %3, $31, $17 \n"
       "call_pal     131 \n"
       "bis $0, $31, %0 \n"
       : "=r" (ret_val)
       : "r" (callno), "r" (key), "r" (value)
       : "$16", "$17", "$18", "$19", "$0"
       );
  return (ret_val);

}

void * __pthread_getspecific(pthread_key_t key)
{
    long ret_val;
    long callno = SYS_aint_pthread_getspecific;

#ifdef DEBUG
  printf("AINT pthread_getspecific: making syscall %d\n", callno); fflush(stdout);
#endif
  asm( "bis %1, $31, $0 \n"
       "bis %2, $31, $16 \n"
       "call_pal     131 \n"
       "bis $0, $31, %0 \n"
       : "=r" (ret_val)
       : "r" (callno), "r" (key)
       : "$16", "$17", "$18", "$19", "$0"
       );
  return ( pthread_t ) ret_val;

}


#ifndef pthread_self
pthread_t
pthread_self()
{
  return __pthread_self();
}
#endif

pthread_t
__pthread_self()
{
    long ret_val;
    long callno = SYS_aint_pthread_self;

#ifdef DEBUG
    printf("AINT pthred_self: making syscall %d\n", callno); fflush(stdout);
#endif
    asm( "bis %1, $31, $0 \n"
	 "call_pal 131 \n"
	 "bis $0, $31, %0 \n"
	 : "=r" (ret_val)
	 : "r" (callno)
	 : "$16", "$17", "$18", "$19", "$0"
	 );
    return ( pthread_t ) ret_val;
}

int
__pthread_equal(pthread_t t1, pthread_t t2)
{
    return (t1 == t2);
}

