/* Assembly language routines for implementing instructions */

#include <iostream.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/signal.h>

#include "aint_types.h"
#include "aint.h"
#include "globals.h"
#include "opcodes.h"
#include "protos.h"

/* icode_ptr
unimplemented_op(inflight_inst_ptr ifi, icode_ptr picode, thread_ptr pthread) */

#ifdef SPECULATIVE_EXCEPTIONS
#undef SPECULATIVE_EXCEPTIONS
#endif


/* All store insts set the f/e bit full. We do the sanity checks in AINT_do_{read,write} */
#define SET_FE_FULL() { *(ifi->get_fe_vector()) |= (ifi->get_fe_mask()); }


OP(unimplemented_op)
{
  /* This may be a speculatively issued inst - cannot cause exception */
#ifdef SPECULATIVE_EXCEPTIONS
  fatal((void *) picode->addr << ": unimplemented operation "
	<< desc_table[picode->opnum].opname << "(" << (void *) picode->instr << ")" << endl);
#endif
  return NULL;
}

/*
icode_ptr
reserved_op(icode_ptr picode, thread_ptr pthread) */
OP(reserved_op)
{
#ifdef SPECULATIVE_EXCEPTIONS
  fatal((void *) picode->addr << ": reserved instruction ("
	<< picode->instr << ")" << endl);
#endif
  return NULL;
}

OP(abort)
{
#ifdef SPECULATIVE_EXCEPTIONS
  fatal((void *) picode->addr << ": invalid index (0) into op function table" << endl);
#endif
  return NULL;
}


OP(gentrap_f)
{
  ifi->runstate = R_SIGNAL;
  ifi->signal = SIGTRAP;
  
  cerr << "GENTRAP: tid=" << pthread->tid << ", pc=0x" << (void *)picode->addr << endl;

  /* this is a hack, because we pass control to the debugger after we commit
   * this instruction, but we want the PC to remain the same until the
   * debugger either replaces the trap instruction or moves us past the
   * trap instruction.  
   */
  return picode->next;
}

OP(rduniq_f)
{
  MapReg(RET_VAL_REG) = pthread->unique;
  return picode->next;
}
OP(wruniq_f)
{
  pthread->unique = MapReg(A0);
  return picode->next;
}
OP(bpt_f)
{
  ifi->runstate = R_SIGNAL;
  ifi->signal = SIGTRAP;

  cerr << "BPT: tid=" << pthread->tid << ", pc=0x" << (void *) picode->addr << endl;
  /* this is a hack, because we pass control to the debugger after we commit
   * this instruction, but we want the PC to remain the same until the
   * debugger either replaces the trap instruction or moves us past the
   * trap instruction.  
   */
  return picode->next;
}

UNIMPLEMENTED(bugchk_f)

/* chmk is the same as callsys */

OP(chmk_f)
{
  return callsys_f(ifi, picode, ppath, pthread);
}

UNIMPLEMENTED(imb_f)
UNIMPLEMENTED(halt_f)
UNIMPLEMENTED(draina_f)
UNIMPLEMENTED(nphalt_f)
UNIMPLEMENTED(cobratt_f)
UNIMPLEMENTED(cserve_f)
UNIMPLEMENTED(ipir_f)
UNIMPLEMENTED(cflush_f)
UNIMPLEMENTED(rti_f)
UNIMPLEMENTED(rtsys_f)
UNIMPLEMENTED(whami_f)
UNIMPLEMENTED(rdusp_f)
UNIMPLEMENTED(wrperfmon_f)
UNIMPLEMENTED(wrusp_f)
UNIMPLEMENTED(wrkgp_f)
UNIMPLEMENTED(rdps_f)
UNIMPLEMENTED(swpipl_f)
UNIMPLEMENTED(wrent_f)
UNIMPLEMENTED(tbi_f)
UNIMPLEMENTED(rdval_f)
UNIMPLEMENTED(wrval_f)
UNIMPLEMENTED(swpctx_f)
UNIMPLEMENTED(jtopal_f)
UNIMPLEMENTED(wrvptptr_f)
UNIMPLEMENTED(wrfen_f)
UNIMPLEMENTED(mtpr_mces_f)

  /* RESERVED(opc01_f) */

RESERVED(opc02_f)
RESERVED(opc03_f)
RESERVED(opc04_f)

  /* RESERVED(opc05_f) */

RESERVED(opc06_f)
RESERVED(opc07_f)

OP(lda_f)
{
#ifdef DEBUG_LDA  
  fprintf(Aint_output, "$%d(%lx) + %lx => ",
	  picode->args[RB], REG(RB), picode->immed);
#endif
  REG(RA) = REG(RB) + picode->immed;
#ifdef DEBUG_LDA  
  fprintf(Aint_output, "$%d(%lx)\n", picode->args[RA], REG(RA));
#endif
  ZERO_ZERO_REG;
  return picode->next;
}

OP(ldah_f)
{
  REG(RA) = REG(RB) + picode->immed;
  ZERO_ZERO_REG;
  return picode->next;
}

RESERVED(opc0a_f)

OP(ldq_u_f)
{
  REG(RA) = (long) ifi->data;
  ZERO_ZERO_REG;
  return picode->next;
}

RESERVED(opc0c_f)
RESERVED(opc0d_f)
RESERVED(opc0e_f)

OP(stq_u_f)
{
  if (ifi->paddr) {
    *(long *) ifi->paddr = ifi->data;
    SET_FE_FULL();
  }
  return picode->next;
}

#define OPFMT(NAME) \
OP(NAME ## _f) \
{ \
  long lv1, rv1, rv2; \
\
  lv1 = REG(RC); \
  rv1 = REG(RA); \
  if (!((union alpha_instruction *) (&picode->instr))->o_format.form) { \
    rv2 = REG(RB);  \
  } else { \
    rv2 = (long) picode->literal; \
  }    \
  /* printf("NAME  =r%d r%di (0x%lx) 0x%lx ", picode->args[RC], */  \
  /* picode->args[RA], rv1, rv2); */  \
  asm(#NAME "	%1, %2, %0" : "&=r" (lv1) : "r" (rv1) , "r" (rv2)); \
  /* if (NAME ## _opn == addl_opn) printf("addl(%lx, %lx) => 0x%lx\n", rv1, rv2, lv1); */\
  REG(RC) = lv1; \
  ZERO_ZERO_REG; \
  return(picode->next); \
}

OPFMT(addl)
OPFMT(s4addl)
OPFMT(subl)
OPFMT(s4subl)
OPFMT(cmpbge)
OPFMT(s8addl)
OPFMT(s8subl)
OPFMT(cmpult)
OPFMT(addq)
OPFMT(s4addq)
OPFMT(subq)
OPFMT(s4subq)
OPFMT(cmpeq)
OPFMT(s8addq)
OPFMT(s8subq)
OPFMT(cmpule)
OPFMT(addlv)
OPFMT(sublv)
OPFMT(cmplt)
OPFMT(addqv)
OPFMT(subqv)
OPFMT(cmple)
OPFMT(and)
OPFMT(bic)

/* OPFMT(cmovlbs) */

OP(cmovlbs_f) {
    long rbv = ((picode->instr & 0x1000)? (picode->literal): REG(RB));
    if (REG(RA)&1) REG(RC) = rbv;
    ZERO_ZERO_REG;
    return picode->next;
}

/* OPFMT(cmovlbc) */
OP(cmovlbc_f) {
    long rbv = ((picode->instr & 0x1000)? (picode->literal): REG(RB));
    if ( (REG(RA)&1) == 0) REG(RC) = rbv;
    ZERO_ZERO_REG;
    return picode->next;
}

OPFMT(bis)

/* OPFMT(cmoveq) */
OP(cmoveq_f) {
    long rbv = ((picode->instr & 0x1000)? (picode->literal): REG(RB));
    if (REG(RA) == 0) REG(RC) = rbv;
    ZERO_ZERO_REG;
    return picode->next;
}

/* OPFMT(cmovne) */
OP(cmovne_f) {
    long rbv = ((picode->instr & 0x1000)? (picode->literal): REG(RB));
    if (REG(RA)) REG(RC) = rbv;
    ZERO_ZERO_REG;
    return picode->next;
}

OPFMT(ornot)
OPFMT(xor)

/* OPFMT(cmovlt) */
OP(cmovlt_f) {
    long rbv = ((picode->instr & 0x1000)? (picode->literal): REG(RB));
    if (REG(RA)<0) REG(RC) = rbv;
    ZERO_ZERO_REG;
    return picode->next;
}

/* OPFMT(cmovge) */
OP(cmovge_f) {
    long rbv = ((picode->instr & 0x1000)? (picode->literal): REG(RB));
    if (REG(RA)>=0) REG(RC) = rbv;
    ZERO_ZERO_REG;
    return picode->next;
}

OPFMT(eqv)

/*OPFMT(cmovle) */
OP(cmovle_f) {
    long rbv = ((picode->instr & 0x1000)? (picode->literal): REG(RB));
    if (REG(RA)<=0) REG(RC) = rbv;
    ZERO_ZERO_REG;
    return picode->next;
}

/* OPFMT(cmovgt) */
OP(cmovgt_f) {
    long rbv = ((picode->instr & 0x1000)? (picode->literal): REG(RB));
    if (REG(RA)>0) REG(RC) = rbv;
    ZERO_ZERO_REG;
    return picode->next;
}

OPFMT(mskbl)
OPFMT(extbl)
OPFMT(insbl)
OPFMT(mskwl)
OPFMT(extwl)
OPFMT(inswl)
OPFMT(mskll)
OPFMT(extll)
OPFMT(insll)
OPFMT(zap)
OPFMT(zapnot)
OPFMT(mskql)
OPFMT(srl)
OPFMT(extql)
OPFMT(sll)
OPFMT(insql)
OPFMT(sra)
OPFMT(mskwh)
OPFMT(inswh)
OPFMT(extwh)
OPFMT(msklh)
OPFMT(inslh)
OPFMT(extlh)
OPFMT(mskqh)
OPFMT(insqh)
OPFMT(extqh)
OPFMT(mull)
OPFMT(mulq)
OPFMT(umulh)
OPFMT(mullv)
OPFMT(mulqv)

RESERVED(opc14_f)

EMPTY_OP(trapb_f)
EMPTY_OP(excb_f)
EMPTY_OP(mb_f)
EMPTY_OP(wmb_f)
EMPTY_OP(fetch_f)
EMPTY_OP(fetch_m_f)

OP(rpcc_f)
{
  REG(RA) = 0; /* (long) ifi->time; */
  ZERO_ZERO_REG;
  return picode->next;
}

UNIMPLEMENTED(rc_f)
UNIMPLEMENTED(rs_f)

RESERVED(pal19_f)

#define JUMP_OP(NAME) \
OP(NAME ## _f) \
{ \
   REG(RA) = IR2V(picode->next); \
   ZERO_ZERO_REG; \
   if (0) cerr << "\t 0x" << (void *) picode->addr << ": " #NAME " 0x" << (void *) REG(RB) << endl; \
   if (0) fprintf(stderr, "\t %lx: " #NAME " 0x%lx\n", picode->addr, REG(RB)); \
   return IV2R(pthread->pprocess, REG(RB)); \
}
 
JUMP_OP(jmp)
JUMP_OP(jsr)
JUMP_OP(ret)
JUMP_OP(jsr_coroutine)
 
RESERVED(pal1b_f)
RESERVED(opc1c_f)
RESERVED(pal1d_f)
RESERVED(pal1e_f)
RESERVED(pal1f_f)

#define FP_RD 1
#define FP_WR 0

/* The following hack fails if the preprocessor recognizes strings and
 * passes them unchanged. Use the -traditional-cpp option with gcc to
 * enable NAME to be substituted in the string
 */

#define MEM_FP(NAME, rw, size) \
OP(NAME ## _f) \
{ \
  if (0) informative(#NAME ": " << (void *) picode->addr << " " << (void *) picode->instr << " " << picode->opnum); \
  /* if (0) informative(#NAME ": %lx %x %d\n"  picode->addr, picode->instr, picode->opnum); */ \
  if (rw) /* This is a read into reg */ {\
   asm(#NAME " %0,%1" : "=f" (FP(RA)) : "o" (* (size *) &ifi->data)); \
   MapFP(ZERO_REGISTER) = 0.0; \
  } else {\
	    /* asm(#NAME " %1,%0" : "=o" (* (size *) pthread->paddr) : "f" (FP(RA))); */ \
    if (ifi->paddr) {\
      *(size *) ifi->paddr = ifi->data; \
      SET_FE_FULL(); \
    } \
  } \
  return (picode->next); \
}

MEM_FP(ldf, FP_RD, long)
MEM_FP(ldg, FP_RD, long)
MEM_FP(lds, FP_RD, long)
MEM_FP(ldt, FP_RD, long)


MEM_FP(stf, FP_WR, int)
MEM_FP(stg, FP_WR, long)
MEM_FP(sts, FP_WR, int)
MEM_FP(stt, FP_WR, long)

void aint_cvt_fp_mem(unsigned int opnum, double fpreg, ulong *memloc)
{
  switch(opnum) {
  case stf_opn:
    asm("stf %1, %0" : "=o" (* memloc) : "f" (fpreg));
    break;
  case stg_opn:
    asm("stg %1, %0" : "=o" (* memloc) : "f" (fpreg));
    break;
  case sts_opn:
    asm("sts %1, %0" : "=o" (* memloc) : "f" (fpreg));
    break;
  case stt_opn:
    asm("stt %1, %0" : "=o" (* memloc) : "f" (fpreg));
    break;
  default:
    break;
  }
}


OP(ldb_f) {
  REG(RA) = (char)ifi->data;
  ZERO_ZERO_REG;
  return picode->next;
}

OP(ldw_f) {
  REG(RA) =(short) ifi->data;
  ZERO_ZERO_REG;
  return picode->next;
}

OP(ldl_f) {
  REG(RA) = ((int)ifi->data);
  ZERO_ZERO_REG;
  return picode->next;
}

OP(ldq_f) {
  REG(RA) = ifi->data;
  ZERO_ZERO_REG;
  return picode->next;
}

OP(ldq_s_f) {
  REG(RA) = ifi->data;
  ZERO_ZERO_REG;
  /* Set f/e bit empty. Sanity checks done in AINT_do_{read,write}  */

  /* fe_mask has a 1-bit in the position corresponding to vaddr */
  *(ifi->get_fe_vector()) = *(ifi->get_fe_vector()) & (~ifi->get_fe_mask());

  return picode->next;
}

OP(ldl_l_f)
{
  /* Load locked.
   * Load data into register
   * Record the address into the locked_addr register, and set lock_flag
   */

  REG(RA) = (int) ifi->data;
  pthread->locked_addr = ifi->paddr >> LOCK_RANGE_SZ;
  pthread->lock_flag = 1; 
  if (0)
    cerr << pthread->tid << " LDL_L: 0x" << (void *) picode->addr << "(" << (void *) ifi->vaddr << ")" << endl;
  ZERO_ZERO_REG;
  return picode->next;
}

OP(ldq_l_f)
{ 
  REG(RA) = (long) ifi->data;
  pthread->locked_addr = ifi->paddr >> LOCK_RANGE_SZ;
  pthread->lock_flag = 1;
  ZERO_ZERO_REG;
  if (0)
    cerr << pthread->tid << " LDQ_L: 0x" << (void *) picode->addr << "(" << (void *) ifi->vaddr << ")" << endl;
  return(picode->next);
}

OP(stb_f)
{
  if (ifi->paddr) {
    *(char *) ifi->paddr = (char) ifi->data;
    SET_FE_FULL();
  }
  return picode->next;
}

OP(stw_f)
{
  if (ifi->paddr) {
    *(short *) ifi->paddr = (short) ifi->data;
    SET_FE_FULL();
  }
  return picode->next;
}

OP(stl_f)
{
  if (ifi->paddr) {
    *(int *) ifi->paddr = (int) ifi->data;
    SET_FE_FULL();
  }
  return picode->next;
}

OP(stq_f)
{
  if (ifi->paddr) {
    * (long *) ifi->paddr = ifi->data;
    SET_FE_FULL();
  }
  return picode->next;
}

OP(stq_s_f)
{
  if (ifi->paddr) {
    * (long *) ifi->paddr = ifi->data;
    SET_FE_FULL();
  }
  return picode->next;
}

OP(stl_c_f)
{
  /* Store conditional
   * If lock flag is set, store RA at address, and clear lock-flags for all
   *   threads which have this address locked.
   * RA gets lock flag and lock flag is set to 0
   */
  int p, rav;
  long lock_range_addr = (ifi->paddr >> LOCK_RANGE_SZ);

  rav = (int) REG(RA);
  REG(RA) = pthread->lock_flag; ZERO_ZERO_REG;

  if (0) {
    cerr << pthread->tid << " STL_C: " << (void *) picode->addr << " (" << (void *) ifi->vaddr << ") "
	 << ((pthread->lock_flag) ? "locked" : "revoked");
  }

  if (pthread->lock_flag) {
    * (int *) ifi->paddr = rav;
    SET_FE_FULL();    /* Strange for stX_c to set f/e bit, but here for completeness */
    for (p=0; p<ThreadCount; p++)
      if (Threads[p]->locked_addr == lock_range_addr)
	Threads[p]->lock_flag = 0;
  }
  return picode->next;
}

OP(stq_c_f)
{
  int p;
  long rav;
  long lock_range_addr = (ifi->paddr >> LOCK_RANGE_SZ);

  rav = REG(RA);
  REG(RA) = pthread->lock_flag; ZERO_ZERO_REG;

  if (0)
    fprintf(stderr, "%d STQ_C: %lx (%lx) %s\n",
	    pthread->tid, picode->addr, ifi->vaddr,
	    (pthread->lock_flag) ? "locked" : "revoked");

  if (pthread->lock_flag) {
    * (long *) ifi->paddr = rav;
    SET_FE_FULL();
    for (p=0; p<ThreadCount; p++)
      if (Threads[p]->locked_addr == lock_range_addr)
	Threads[p]->lock_flag = 0;
  }
    
  return picode->next;
}


OP(br_f) 
{ 
  REG(RA) = IR2V(picode->next);
  ZERO_ZERO_REG;
  return picode->target; 
}

OP(bsr_f)
{ 
  if (0) fprintf(stderr, "%lx:  BSR 0x%lx\n", picode->addr, IR2V(picode->target));
  REG(RA) = IR2V(picode->next);
  ZERO_ZERO_REG;
  return picode->target; 
}

#define FP_BR(NAME, LABEL) \
OP(NAME ## _f) \
{ \
  long i=1, cond = 1; \
  asm (#NAME " %1, " # LABEL "; subq %2,1,%0; " #LABEL ":" : "=r" (cond) : \
       "f" (FP(RA)) , "r" (i)); \
  if (cond) { \
    return picode->target; \
  } else { \
    return picode->next; \
  } \
}

FP_BR(fbeq, lfbeq)
FP_BR(fblt, lfblt)
FP_BR(fble, lfble)

FP_BR(fbne, lfbne)
FP_BR(fbge, lfbge)
FP_BR(fbgt, lfbgt)

OP(blbc_f) { 
  if (REG(RA)&1) {
    return picode->next; 
  } else {
    return picode->target;
  }
}

OP(beq_f) { 
  if (REG(RA)) {
    return picode->next; 
  }
  return picode->target;
}

OP(blt_f) {
  if (REG(RA) >= 0) {
    return picode->next; 
  } 
  return picode->target;
}

OP(ble_f) {
  if (REG(RA) > 0) {
    return picode->next;
  }
  return picode->target;
}

OP(blbs_f) {
  if (REG(RA)&1) {
    return picode->target;
  }
  return picode->next;
}

OP(bne_f) {
  if (REG(RA)) {
    return picode->target;
  }
  return picode->next;
}

OP(bge_f) { 
  if (REG(RA) >= 0) {
    return picode->target; 
  }
  return picode->next; 
}

OP(bgt_f) {
  if (REG(RA) > 0) {
    return picode->target;
  }
  return picode->next;
}

double get_fpcr()
{
  double fpcr; asm ("trapb; mf_fpcr %0; trapb"
         : "=f" (fpcr));
  return fpcr;
}

void put_fpcr(double fpcr)
{
  asm ("trapb; mt_fpcr %0; trapb"
       : : "f" (fpcr));
}
