#include <stdio.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include "alpha_lock.h"

int main(int argc, char *argv[])
{
  /* Allocate shmem, fork 2 processes, and do several thousand locked reads
   * and writes. Wait for child to die, then print results
   */
  key_t k;
  long shmid;
  int i, n, s, pid, Nprocs;
  long *seg, x, mysum;
  long tmp;
  int whoami;
  struct shmid_ds buf;

  fprintf(stderr,"Usage: locks nprocs nloops seed>\n");

  if (argc != 4) exit (-1);

  sscanf(argv[1], "%d", &Nprocs);
  sscanf(argv[2], "%d", &n);
  sscanf(argv[3], "%d", &s);
  srandom(s);

  k = ftok("/vmunix", getpid());

  shmid = shmget(k, sizeof(long)*1024*2, IPC_CREAT|0666);
  seg = shmat(shmid, 0, 0);
  pid = getpid();
  for (i=0; i<Nprocs; i++) {
    if (getpid() == pid) fork();
  }

  whoami = getpid();

  if (whoami == pid) {
    for (i=0; i<Nprocs; i++)
      wait(&tmp);
    /* Print the global sum */
    printf("All children done. Shared sum is %ld\n", seg[1024]);
    shmctl(shmid, IPC_RMID, &buf);
    exit(0);
  }

  mysum = 0;
  
  printf("%d : Begining long sequence of additions.\n", whoami);

  /* Do a long sequence of additions. */
  for(i=0; i<n; i++) {
    tmp = random()>>24;
    mysum += tmp;
    /* Do an atomic update */
#if 0
    asm("_again2: ldl_q %0 %1; addq %0 %2 %0; stq_c %0 %1; beq %0, _again2":
	"&=r" (x), "=o"(*seg): "r"(tmp));
#else
    alpha_atomic_add_q(seg+1024, tmp);
#endif
    alpha_mb();
  }
  printf("%d: mysum %d, shared sum %ld\n", whoami, mysum, seg[1024]);
  /* Detach mem seg */
  shmdt(seg);

  exit (0);
}
